# -*- coding: utf-8 -*-
#16-6-18
# create by: snower

import sys
import greenlet
import functools
from tornado.ioloop import IOLoop
from tornado.concurrent import is_future, Future


class HandlerException(Exception):
    def __init__(self, name, e, exc_info):
        super(HandlerException, self).__init__(e.message)

        self.name = name
        self.e = e
        self.exc_info = exc_info


class HandlerWrapper(object):
    def __init__(self, server, processor, handler):
        self._server = server
        self._processor = processor
        self._handler = handler

    def __getattr__(self, name):
        if name not in self._processor._processMap:
            return getattr(self._handler, name)

        func = getattr(self._handler, name)
        ioloop = IOLoop.current()

        def run(child_gr, *args, **kwargs):
            try:
                result = func(*args, **kwargs)
                if not is_future(result):
                    return child_gr.switch(result)
            except:
                exc_info = sys.exc_info()
                result = Future()
                result.set_exc_info(exc_info)
            return ioloop.add_future(result, child_gr.switch)

        @functools.wraps(func)
        def _(*args, **kwargs):
            self._server.handing_count += 1
            try:
                child_gr = greenlet.getcurrent()
                main = child_gr.parent
                if main is None:
                    return func(*args, **kwargs)

                ioloop.add_callback(run, child_gr, *args, **kwargs)
                result = main.switch()
                if is_future(result):
                    result = result.result()
                return result
            except Exception as e:
                raise HandlerException(name, e, sys.exc_info())
            finally:
                self._server.handing_count -= 1

        setattr(self, name, _)
        return _


# -*- coding: utf-8 -*-
#14-6-11
# create by: snower

import sys
import logging
import socket
import errno
import ssl
import greenlet
import traceback
from tornado import tcpserver
from tornado.concurrent import TracebackFuture
from tornado.tcpserver import ssl_wrap_socket, app_log
from tornado.iostream import IOStream as BaseIOStream, SSLIOStream, StreamClosedError, errno_from_exception, _ERRNO_WOULDBLOCK
from thrift.transport import TTransport
from thrift.Thrift import TApplicationException as BaseTApplicationException, TMessageType
from .handler import HandlerWrapper, HandlerException


class TApplicationException(BaseTApplicationException):
    HANDLER_UNKNOWN_ERROR = 1024


class IOStream(BaseIOStream):
    def __init__(self, socket, *args, **kwargs):
        super(IOStream, self).__init__(socket, *args, **kwargs)

        if socket and self._state is None:
            self._state = self.io_loop.ERROR | self.io_loop.READ
            self.io_loop.add_handler(self.fileno(), self._handle_events, self._state)

    def _handle_events(self, fd, events):
        if self._closed:
            return
        try:
            if events & self.io_loop.READ:
                self._handle_read()
            if self._closed:
                return
            if events & self.io_loop.WRITE:
                self._handle_write()
            if self._closed:
                return
            if events & self.io_loop.ERROR:
                self.error = self.get_fd_error()
                self.io_loop.add_callback(self.close)
                return
        except Exception:
            self.close(exc_info=True)
            raise

    def _handle_read(self):
        chunk = True

        while True:
            try:
                chunk = self.socket.recv(self.read_chunk_size)
                if not chunk:
                    break
                self._read_buffer.append(chunk)
                self._read_buffer_size += len(chunk)
            except (socket.error, IOError, OSError) as e:
                en = e.errno if hasattr(e, 'errno') else e.args[0]
                if en in _ERRNO_WOULDBLOCK:
                    break

                if en == errno.EINTR:
                    continue

                self.close(exc_info=True)
                return

        if self._read_future is not None and self._read_buffer_size >= self._read_bytes:
            future, self._read_future = self._read_future, None
            data = b"".join(self._read_buffer)
            self._read_buffer.clear()
            self._read_buffer_size = 0
            self._read_bytes = 0
            future.set_result(data)

        if not chunk:
            self.close()
            return

    def read(self, num_bytes):
        assert self._read_future is None, "Already reading"
        if self._closed:
            raise StreamClosedError(real_error=self.error)

        future = self._read_future = TracebackFuture()
        self._read_bytes = num_bytes
        self._read_partial = False
        if self._read_buffer_size >= self._read_bytes:
            future, self._read_future = self._read_future, None
            data = b"".join(self._read_buffer)
            self._read_buffer.clear()
            self._read_buffer_size = 0
            self._read_bytes = 0
            future.set_result(data)
        return future

    def _handle_write(self):
        while self._write_buffer:
            try:
                data = self._write_buffer.popleft()
                num_bytes = self.socket.send(data)
                self._write_buffer_size -= num_bytes
                if num_bytes < len(data):
                    self._write_buffer.appendleft(data[num_bytes:])
                    return
            except (socket.error, IOError, OSError) as e:
                en = e.errno if hasattr(e, 'errno') else e.args[0]
                if en in _ERRNO_WOULDBLOCK:
                    self._write_buffer.appendleft(data)
                    break

                self.close(exc_info=True)
                return

        if not self._write_buffer:
            if self._state & self.io_loop.WRITE:
                self._state = self._state & ~self.io_loop.WRITE
                self.io_loop.update_handler(self.fileno(), self._state)
            if self._write_future:
                future, self._write_future = self._write_future, None
                future.set_result(None)

    def write(self, data):
        assert isinstance(data, bytes)
        if self._closed:
            raise StreamClosedError(real_error=self.error)

        if not data:
            future = self._write_future if self._write_future else TracebackFuture()
            future.set_result(None)
            return future

        self._write_buffer.append(data)
        self._write_buffer_size += len(data)
        future = self._write_future = TracebackFuture()

        self._handle_write()
        if self._write_buffer:
            if not self._state & self.io_loop.WRITE:
                self._state = self._state | self.io_loop.WRITE
                self.io_loop.update_handler(self.fileno(), self._state)

        return future


class TTornadoServer(tcpserver.TCPServer):
    def __init__(self, processor, input_transport_factory, input_protocol_factory, output_transport_factory = None, output_protocol_factory = None, *args, **kwargs):
        super(TTornadoServer,self).__init__(*args, **kwargs)

        self.processor = processor
        self.processor_count = 0
        self.handing_count = 0
        self.stoped = False
        self.input_transport_factory = input_transport_factory
        self.output_transport_factory = output_transport_factory
        self.input_protocol_factory = input_protocol_factory
        self.output_protocol_factory = output_protocol_factory

        self.processor._handler = HandlerWrapper(self, self.processor, self.processor._handler)

    def handle_exception(self, exc_info):
        logging.error("processor error: %s", exc_info = exc_info)

    def format_exception_message(self, e, exc_info):
        etype, value, tb = exc_info
        return "%s:%s\n%s" % (e.__class__.__name__, str(e), ''.join(traceback.format_exception(etype, value, tb, None)))

    def handle_handler_exception(self, e, oprot):
        x = TApplicationException(TApplicationException.HANDLER_UNKNOWN_ERROR, self.format_exception_message(e.e, e.exc_info))
        oprot.writeMessageBegin(e.name, TMessageType.EXCEPTION, 0)
        x.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process(self, itrans, otrans):
        iprot = self.input_protocol_factory.getProtocol(itrans)
        oprot = self.output_protocol_factory.getProtocol(otrans) if otrans and self.output_protocol_factory else iprot

        try:
            self.processor_count += 1
            while not self.stoped:
                try:
                    self.processor.process(iprot, oprot)
                except HandlerException as e:
                    self.handle_handler_exception(e, oprot)
        except (TTransport.TTransportException, IOError, StreamClosedError, EOFError):
            pass
        except Exception:
            self.handle_exception(sys.exc_info())
        finally:
            self.processor_count -= 1

        itrans.close()
        if otrans:
            otrans.close()
        if self.handing_count == 0 and self.stoped:
            stoped_future, self.stoped = self.stoped, None
            stoped_future.set_result(None)

    def _handle_connection(self, connection, address):
        if self.ssl_options is not None:
            assert ssl, "Python 2.6+ and OpenSSL required for SSL"
            try:
                connection = ssl_wrap_socket(connection,
                                             self.ssl_options,
                                             server_side=True,
                                             do_handshake_on_connect=False)
            except ssl.SSLError as err:
                if err.args[0] == ssl.SSL_ERROR_EOF:
                    return connection.close()
                else:
                    raise
            except socket.error as err:
                if errno_from_exception(err) in (errno.ECONNABORTED, errno.EINVAL):
                    return connection.close()
                else:
                    raise
        try:
            if self.ssl_options is not None:
                stream = SSLIOStream(connection, io_loop=self.io_loop,
                                     max_buffer_size=self.max_buffer_size,
                                     read_chunk_size=self.read_chunk_size)
            else:
                stream = IOStream(connection, io_loop=self.io_loop,
                                  max_buffer_size=self.max_buffer_size,
                                  read_chunk_size=self.read_chunk_size)
            self.handle_stream(stream, address)
        except Exception:
            self.handle_exception(sys.exc_info())

    def handle_stream(self, stream, address):
        itrans = self.input_transport_factory.getTransport(stream)
        otrans = self.output_transport_factory.getTransport(stream) if self.output_transport_factory else None
        child_gr = greenlet.greenlet(self.process)
        child_gr.switch(itrans, otrans)

    def stop(self):
        super(TTornadoServer, self).stop()

        future = self.stoped = TracebackFuture()
        if self.handing_count <= 1:
            self.stoped = None
            future.set_result(None)
        return future


# -*- coding: utf-8 -*-
# 14-8-16
# create by: snower

import time
import socket
import errno
from tornado.concurrent import TracebackFuture
from tornado.iostream import IOStream, StreamClosedError, _ERRNO_WOULDBLOCK


class TStreamConnectTimeoutError(Exception):
    pass


class TStream(IOStream):
    def __init__(self, host='127.0.0.1', port=9090, unix_socket=None, socket_family=socket.AF_UNSPEC, timeout=5):
        self._host = host
        self._port = port
        self._unix_socket = unix_socket
        self._socket_family = socket_family
        self._timeout = timeout

        if self._unix_socket:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        else:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

        super(TStream, self).__init__(sock)

    def open(self):
        if self._unix_socket:
            address = self._unix_socket
        else:
            address = (self._host, self._port)
        future = self.connect(address)
        if self._timeout > 0:
            def timeout():
                if self._connecting:
                    self.close((None, TStreamConnectTimeoutError(), None))
            self.io_loop.add_timeout(time.time() + self._timeout, timeout)
        return future

    def _handle_events(self, fd, events):
        if self._closed:
            return
        try:
            if self._connecting:
                self._handle_connect()
            if self._closed:
                return
            if events & self.io_loop.READ:
                self._handle_read()
            if self._closed:
                return
            if events & self.io_loop.WRITE:
                self._handle_write()
            if self._closed:
                return
            if events & self.io_loop.ERROR:
                self.error = self.get_fd_error()
                self.io_loop.add_callback(self.close)
                return
        except Exception:
            self.close(exc_info=True)
            raise

    def _handle_connect(self):
        super(TStream, self)._handle_connect()

        if not self.closed():
            self._state = self.io_loop.ERROR | self.io_loop.READ
            if self._write_buffer:
                self._state = self._state | self.io_loop.WRITE
            self.io_loop.update_handler(self.fileno(), self._state)

    def _handle_read(self):
        chunk = True

        while True:
            try:
                chunk = self.socket.recv(self.read_chunk_size)
                if not chunk:
                    break
                self._read_buffer.append(chunk)
                self._read_buffer_size += len(chunk)
            except (socket.error, IOError, OSError) as e:
                en = e.errno if hasattr(e, 'errno') else e.args[0]
                if en in _ERRNO_WOULDBLOCK:
                    break

                if en == errno.EINTR:
                    continue

                self.close(exc_info=True)
                return

        if self._read_future is not None and self._read_buffer_size >= self._read_bytes:
            future, self._read_future = self._read_future, None
            data = b"".join(self._read_buffer)
            self._read_buffer.clear()
            self._read_buffer_size = 0
            self._read_bytes = 0
            future.set_result(data)

        if not chunk:
            self.close()
            return

    def read(self, num_bytes):
        assert self._read_future is None, "Already reading"
        if self._closed:
            raise StreamClosedError(real_error=self.error)

        future = self._read_future = TracebackFuture()
        self._read_bytes = num_bytes
        self._read_partial = False
        if self._read_buffer_size >= self._read_bytes:
            future, self._read_future = self._read_future, None
            data = b"".join(self._read_buffer)
            self._read_buffer.clear()
            self._read_buffer_size = 0
            self._read_bytes = 0
            future.set_result(data)
        return future

    def _handle_write(self):
        while self._write_buffer:
            try:
                data = self._write_buffer.popleft()
                num_bytes = self.socket.send(data)
                self._write_buffer_size -= num_bytes
                if num_bytes < len(data):
                    self._write_buffer.appendleft(data[num_bytes:])
                    return
            except (socket.error, IOError, OSError) as e:
                en = e.errno if hasattr(e, 'errno') else e.args[0]
                if en in _ERRNO_WOULDBLOCK:
                    self._write_buffer.appendleft(data)
                    break

                self.close(exc_info=True)
                return

        if not self._write_buffer:
            if self._state & self.io_loop.WRITE:
                self._state = self._state & ~self.io_loop.WRITE
                self.io_loop.update_handler(self.fileno(), self._state)
            if self._write_future:
                future, self._write_future = self._write_future, None
                future.set_result(None)

    def write(self, data):
        assert isinstance(data, bytes)
        if self._closed:
            raise StreamClosedError(real_error=self.error)

        if not data:
            future = self._write_future if self._write_future else TracebackFuture()
            future.set_result(None)
            return future

        self._write_buffer.append(data)
        self._write_buffer_size += len(data)
        future = self._write_future = TracebackFuture()

        if not self._connecting:
            self._handle_write()
            if self._write_buffer:
                if not self._state & self.io_loop.WRITE:
                    self._state = self._state | self.io_loop.WRITE
                    self.io_loop.update_handler(self.fileno(), self._state)

        return future


# -*- coding: utf-8 -*-
# 14-8-19
# create by: snower

import sys
import greenlet
from tornado import gen
from tornado.concurrent import TracebackFuture
from tornado.ioloop import IOLoop
from .transport.transport import TIOStreamTransportFactory


def async_call_method(fun, *args, **kwargs):
    future = TracebackFuture()
    def finish():
        try:
            result = fun(*args, **kwargs)
            if future._callbacks:
                IOLoop.current().add_callback(future.set_result, result)
            else:
                future.set_result(result)
        except:
            if future._callbacks:
                IOLoop.current().add_callback(future.set_exc_info, sys.exc_info())
            else:
                future.set_exc_info(sys.exc_info())
    child_gr = greenlet.greenlet(finish)
    child_gr.switch()
    return future


class PoolClient(object):
    def __init__(self, client_class, itrans_pool, pfactory, tfactory = TIOStreamTransportFactory(), otrans_pool=None):
        self._client_class = client_class
        self._itrans_pool = itrans_pool
        self._otrans_pool = otrans_pool
        self._tfactory = tfactory
        self._pfactory = pfactory

    def __getattr__(self, name):
        @gen.coroutine
        def _(*args, **kwargs):
            istream = yield self._itrans_pool.get_stream()
            ostream = (yield self._otrans_pool.get_stream()) if self._otrans_pool else None
            iprot = self._pfactory.getProtocol(self._tfactory.getTransport(istream))
            oprot = self._pfactory.getProtocol(self._tfactory.getTransport(ostream)) if ostream else None
            client = self._client_class(iprot, oprot)
            try:
                result = yield async_call_method(getattr(client, name), *args, **kwargs)
            finally:
                self._itrans_pool.release_stream(istream)
                if ostream:
                    self._otrans_pool.release_stream(ostream)
            raise gen.Return(result)
        setattr(self, name, _)
        return _



# -*- coding: utf-8 -*-
#14-6-11
# create by: snower

from setuptools import setup

setup(
    name='torthrift',
    version='0.1.8',
    packages=['torthrift','torthrift.server','torthrift.transport'],
    package_data={
        '': ['README.rst'],
    },
    install_requires=['thrift==0.9.3','tornado>=3.2.1', 'greenlet>=0.4.2',],
    author='snower',
    author_email='sujian199@gmail.com',
    url='http://github.com/snower/TorThrift',
    license='MIT',
    description='Tornado asynchronous thrift server and client',
    long_description='torthrift - presents a Tornado Future-based API and greenlet for non-blocking thrift server and client.'
)
